# Copyright (c) 2021 Ultimaker B.V.
# libCharon is released under the terms of the LGPLv3 or higher.

import socket
import struct
from io import BytesIO, SEEK_SET, SEEK_CUR

from typing import Any, Dict, IO, Optional, List

from Charon.filetypes.GCodeFile import GCodeFile
from urllib.parse import urlparse


## This class is used to read GCode stream that are served
#  dynamically over a TCP connection.
class SocketFileStream(BytesIO):
    def __init__(self, sock_object: socket.socket) -> None:
        super().__init__()
        self.current_line = 0
        self.__socket = sock_object

    def seekable(self) -> bool:
        return True

    def seek(self, offset: int, whence: Optional[int] = None) -> int:
        if whence is None or whence == SEEK_SET:
            self.current_line = offset
        elif whence == SEEK_CUR:
            self.current_line += offset
        else:
            raise ValueError('Unsupported whence mode in seek: %d' % whence)
        return offset

    def readline(self, _size: int = -1) -> bytes:
        self.__socket.send(struct.pack('>I', self.current_line))
        line = b''
        char = b''

        while char != b'\n':
            char = self.__socket.recv(1)
            line += char

        self.current_line += 1
        return line

    def read(self, _size: int = -1) -> bytes:
        raise NotImplementedError("Only readline has been implemented")

    def readlines(self, _hint: int = -1) -> List[bytes]:
        raise NotImplementedError("Only readline has been implemented")

    def tell(self) -> int:
        raise NotImplementedError("Only readline has been implemented")

    def close(self) -> None:
        self.__socket.close()

    def __iter__(self):
        return self

    def __next__(self):
        return self.readline()


class GCodeSocket(GCodeFile):
    mime_type = "text/x-gcode-socket"

    MaximumHeaderLength = 100

    def __init__(self) -> None:
        super().__init__()
        self.__stream = None  # type: Optional[IO[bytes]]
        self.__metadata = {}  # type: Dict[str, Any]
        self.__sock = None

    @staticmethod
    def stream_handler(path: str, mode: str) -> IO:
        url = urlparse(path)
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.connect((url.hostname, 1337))
        return SocketFileStream(sock)



# Copyright (c) 2018 Ultimaker B.V.
# libCharon is released under the terms of the LGPLv3 or higher.
from collections import OrderedDict  # To specify the aliases in order.
from io import BytesIO
import json  # The metadata format.
import re  # To find the path aliases.
from typing import Any, Dict, List, IO, Optional
import xml.etree.ElementTree as ET  # For writing XML manifest files.
import zipfile

from Charon.FileInterface import FileInterface  # The interface we're implementing.
from Charon.OpenMode import OpenMode  # To detect whether we want to read and/or write to the file.
from Charon.ReadOnlyError import ReadOnlyError  # To be thrown when trying to write while in read-only mode.
from Charon.WriteOnlyError import WriteOnlyError  # To be thrown when trying to read while in write-only mode.
from Charon.filetypes.GCodeFile import GCodeFile  # Required for fallback G-Code header parsing.


##  A container file type that contains multiple 3D-printing related files that
#   belong together.
class OpenPackagingConvention(FileInterface):
    # Some constants related to this format.
    _xml_header = ET.ProcessingInstruction("xml",
                                           "version=\"1.0\" encoding=\"UTF-8\"")  # Header element being put atop every XML file.
    _content_types_file = "/[Content_Types].xml"  # Where the content types file is.
    _global_metadata_file = "/Metadata/OPC_Global.json"  # Where the global metadata file is.
    _opc_metadata_relationship_type = "http://schemas.ultimaker.org/package/2018/relationships/opc_metadata"  # Unique identifier of the relationship type that relates OPC metadata to files.
    _metadata_prefix = "/metadata"
    _aliases = OrderedDict([])  # type: Dict[str, str]  # A standard OPC file doest not have default aliases. These must be implemented in inherited classes.

    mime_type = "application/x-opc"

    ##  Initialises the fields of this class.
    def __init__(self) -> None:
        self._mode = None  # type: Optional[OpenMode]        # Whether we're in read or write mode.
        self._stream = None  # type: Optional[IO[bytes]]       # The currently open stream.
        self._zipfile = None  # type: Optional[zipfile.ZipFile] # The zip interface to the currently open stream.
        self._metadata = {}  # type: Dict[str, Any]            # The metadata in the currently open file.
        self._content_types_element = None  # type: Optional[ET.Element] # An XML element holding all the content types.
        self._relations = {}  # type: Dict[str, ET.Element]     # For each virtual path, a relations XML element (which is left out of the file if empty).
        self._open_bytes_streams = {}  # type: Dict[str, IO[bytes]] # With old Python versions, the currently open BytesIO streams that need to be flushed, by their virtual path.

        # The zipfile module may only have one write stream open at a time. So when you open a new stream, close the previous one.
        self._last_open_path = None  # type: Optional[str]
        self._last_open_stream = None  # type: Optional[IO[bytes]]

    def openStream(self, stream: IO[bytes], mime: str = "application/x-opc",
                   mode: OpenMode = OpenMode.ReadOnly) -> None:
        self._mode = mode
        self._stream = stream  # A copy in case we need to rewind for toByteArray. We should mostly be reading via self._zipfile.
        self._zipfile = zipfile.ZipFile(self._stream, self._mode.value, compression=zipfile.ZIP_DEFLATED)

        self._readContentTypes()  # Load or create the content types element.
        self._readRels()  # Load or create the relations.
        self._readMetadata()  # Load the metadata, if any.

    def close(self) -> None:
        if not self._stream:
            raise ValueError("This file is already closed.")
        if self._zipfile is None:
            return

        self.flush()
        self._zipfile.close()

    def flush(self) -> None:
        if not self._stream:
            raise ValueError("Can't flush a closed file.")
        assert self._zipfile is not None

        if self._mode == OpenMode.ReadOnly:
            return  # No need to flush reading of zip archives as they are blocking calls.

        if self._last_open_stream is not None and self._last_open_path not in self._open_bytes_streams:
            self._last_open_stream.close()

        # If using old Python versions (<= 3.5), the write streams were kept in memory to be written all at once when flushing.
        for virtual_path, stream in self._open_bytes_streams.items():
            stream.seek(0)
            self._zipfile.writestr(virtual_path, stream.read())
            stream.close()

        self._writeMetadata()  # Metadata must be updated first, because that adds rels and a content type.
        self._writeContentTypes()
        self._writeRels()

    def listPaths(self) -> List[str]:
        if not self._stream:
            raise ValueError("Can't list the paths in a closed file.")
        paths = [self._zipNameToVirtualPath(zip_name) for zip_name in self._zipfile.namelist()]
        return list(self._metadata.keys()) + paths

    def getData(self, virtual_path: str) -> Dict[str, Any]:
        if not self._stream:
            raise ValueError("Can't get data from a closed file.")
        assert self._zipfile is not None

        if self._mode == OpenMode.WriteOnly:
            raise WriteOnlyError(virtual_path)

        result = {}  # type: Dict[str, Any]
        if virtual_path.startswith(self._metadata_prefix):
            result = self.getMetadata(virtual_path[len(self._metadata_prefix):])
        else:
            canonical_path = self._processAliases(virtual_path)
            if self._resourceExists(canonical_path):
                result[virtual_path] = self.getStream(
                    canonical_path).read()  # In case of a name clash, the file wins. But that shouldn't be possible.

        return result

    def setData(self, data: Dict[str, Any]) -> None:
        if not self._stream:
            raise ValueError("Can't change the data in a closed file.")
        if self._mode == OpenMode.ReadOnly:
            raise ReadOnlyError()
        for virtual_path, value in data.items():
            if virtual_path.startswith(
                    self._metadata_prefix):  # Detect metadata by virtue of being in the Metadata folder.
                self.setMetadata({virtual_path: value[len(self._metadata_prefix):]})
            else:  # Virtual file resources.
                self.getStream(virtual_path).write(value)

    def getMetadata(self, virtual_path: str) -> Dict[str, Any]:
        if not self._stream:
            raise ValueError("Can't get metadata from a closed file.")
        assert self._zipfile is not None

        if self._mode == OpenMode.WriteOnly:
            raise WriteOnlyError(virtual_path)
        canonical_path = self._processAliases(virtual_path)

        # Find all metadata that begins with the specified virtual path!
        result = {}

        if canonical_path in self._metadata:  # The exact match.
            result[self._metadata_prefix + virtual_path] = self._metadata[canonical_path]
        for entry_path, value in self._metadata.items():
            # We only want to match subdirectories of the provided virtual paths.
            # So if you provide "/foo" then we don't want to match on "/foobar"
            # but we do want to match on "/foo/zoo". This is why we check if they
            # start with the provided virtual path plus a slash.
            if entry_path.startswith(canonical_path + "/"):
                # We need to return the originally requested alias, so replace the canonical path with the virtual path.
                result[self._metadata_prefix + virtual_path + "/" + entry_path[len(canonical_path) + 1:]] = value

        # If requesting the size of a file.
        if canonical_path.endswith("/size"):
            requested_resource = canonical_path[:-len("/size")]
            if self._resourceExists(requested_resource):
                result[self._metadata_prefix + virtual_path] = self._zipfile.getinfo(
                    requested_resource.strip("/")).file_size

        return result

    def setMetadata(self, metadata: Dict[str, Any]) -> None:
        if not self._stream:
            raise ValueError("Can't change metadata in a closed file.")
        if self._mode == OpenMode.ReadOnly:
            raise ReadOnlyError()
        metadata = {self._processAliases(virtual_path): metadata[virtual_path] for virtual_path in metadata}
        self._metadata.update(metadata)

    def getStream(self, virtual_path: str) -> IO[bytes]:
        if not self._stream:
            raise ValueError("Can't get a stream from a closed file.")
        assert self._zipfile is not None
        assert self._mode is not None

        if virtual_path.startswith("/_rels"):
            raise OPCError("Writing directly to a relationship file is forbidden.")

        if virtual_path.startswith(self._metadata_prefix):
            return BytesIO(json.dumps(self.getMetadata(virtual_path[len(self._metadata_prefix):])).encode("UTF-8"))

        virtual_path = self._processAliases(virtual_path)
        if not self._resourceExists(virtual_path) and self._mode == OpenMode.ReadOnly:  # In write-only mode, create a new file instead of reading metadata.
            raise FileNotFoundError(virtual_path)

        # The zipfile module may only have one write stream open at a time. So when you open a new stream, close the previous one.
        if self._last_open_stream is not None and self._last_open_path not in self._open_bytes_streams:  # Don't close streams that we still need to flush.
            self._last_open_stream.close()

        # If we are requesting a stream of an image resized, resize the image and return that.
        if self._mode == OpenMode.ReadOnly and ".png/" in virtual_path:
            png_file = virtual_path[:virtual_path.find(".png/") + 4]
            size_spec = virtual_path[virtual_path.find(".png/") + 5:]
            if re.match(r"^\s*\d+\s*x\s*\d+\s*$", size_spec):
                dimensions = []
                for dimension in re.finditer(r"\d+", size_spec):
                    dimensions.append(int(dimension.group()))
                return self._resizeImage(png_file, dimensions[0], dimensions[1])

        self._last_open_path = virtual_path
        try:  # If it happens to match some existing PNG file, we have to rescale that file and return the result.
            self._last_open_stream = self._zipfile.open(virtual_path, self._mode.value)
        except RuntimeError:  # Python 3.5 and before couldn't open resources in the archive in write mode.
            self._last_open_stream = BytesIO()
            self._open_bytes_streams[virtual_path] = self._last_open_stream  # Save this for flushing later.
        return self._last_open_stream

    def toByteArray(self, offset: int = 0, count: int = -1) -> bytes:
        if not self._stream:
            raise ValueError("Can't get the bytes from a closed file.")
        if self._mode == OpenMode.WriteOnly:
            raise WriteOnlyError()
        assert self._zipfile is not None
        assert self._mode is not None

        self._zipfile.close()  # Close the zipfile first so that we won't be messing with the stream without its consent.

        self._stream.seek(offset)
        result = self._stream.read(count)

        self._zipfile = zipfile.ZipFile(self._stream, self._mode.value, compression=zipfile.ZIP_DEFLATED)
        return result

    ##  Adds a new content type to the archive.
    #   \param extension The file extension of the type
    def addContentType(self, extension: str, mime_type: str) -> None:
        if not self._stream:
            raise ValueError("Can't add a content type to a closed file.")
        if self._mode == OpenMode.ReadOnly:
            raise ReadOnlyError()
        assert self._content_types_element is not None

        # First check if it already exists.
        for content_type in self._content_types_element.iterfind("Default"):
            if "Extension" in content_type.attrib and content_type.attrib["Extension"] == extension:
                raise OPCError("Content type for extension {extension} already exists.".format(extension=extension))

        ET.SubElement(self._content_types_element, "Default", Extension=extension, ContentType=mime_type)

    ##  Adds a relation concerning a file type.
    #   \param virtual_path The target file that the relation is about.
    #   \param relation_type The type of the relation. Any reader of OPC should
    #   be able to understand all types that are added via relations.
    #   \param origin The origin of the relation. If the relation concerns a
    #   specific directory or specific file, then you should point to the
    #   virtual path of that file here.
    def addRelation(self, virtual_path: str, relation_type: str, origin: str = "") -> None:
        if not self._stream:
            raise ValueError("Can't add a relation to a closed file.")
        if self._mode == OpenMode.ReadOnly:
            raise ReadOnlyError(virtual_path)
        virtual_path = self._processAliases(virtual_path)

        # First check if it already exists.
        if origin not in self._relations:
            self._relations[origin] = ET.Element("Relationships",
                                                 xmlns="http://schemas.openxmlformats.org/package/2006/relationships")
        else:
            for relationship in self._relations[origin].iterfind("Relationship"):
                if "Target" in relationship.attrib and relationship.attrib["Target"] == virtual_path:
                    raise OPCError("Relation for virtual path {target} already exists.".format(target=virtual_path))

        # Find a unique name.
        unique_id = 0
        while True:
            for relationship in self._relations[origin].iterfind("Relationship"):
                if "Id" in relationship.attrib and relationship.attrib["Id"] == "rel" + str(unique_id):
                    break
            else:  # Unique ID didn't exist yet! It's safe to use
                break
            unique_id += 1
        unique_name = "rel" + str(unique_id)

        # Create the element itself.
        ET.SubElement(self._relations[origin], "Relationship", Target=virtual_path, Type=relation_type, Id=unique_name)

    ##  Figures out if a resource exists in the archive.
    #
    #   This will not match on metadata, only on normal resources.
    #   \param virtual_path: The path to test for.
    #   \return ``True`` if it exists as a normal resource, or ``False`` if it
    #   doesn't.
    def _resourceExists(self, virtual_path: str) -> bool:
        assert self._zipfile is not None

        for zip_name in self._zipfile.namelist():
            zip_virtual_path = self._zipNameToVirtualPath(zip_name)
            if virtual_path == zip_virtual_path:
                return True
            if zip_virtual_path.endswith(".png") and virtual_path.startswith(
                    zip_virtual_path + "/"):  # We can rescale PNG images if you want.
                if re.match(r"^\s*\d+\s*x\s*\d+\s*$", virtual_path[len(
                        zip_virtual_path) + 1:]):  # Matches the form "NxM" with optional whitespace.
                    return True
        return False

    ##  Dereference the aliases for OPC files.
    #
    #   This also adds a slash in front of every virtual path if it has no slash
    #   yet, to allow referencing virtual paths with or without the initial
    #   slash.
    def _processAliases(self, virtual_path: str) -> str:
        if not virtual_path.startswith("/"):
            virtual_path = "/" + virtual_path

        # Replace all aliases.
        for regex, replacement in self._aliases.items():
            if regex.startswith("/"):
                expression = r"^" + regex
            else:
                expression = regex
            virtual_path = re.sub(expression, replacement, virtual_path)

        return virtual_path

    ##  Convert the resource name inside the zip to a virtual path as this
    #   library specifies it should be.
    #   \param zip_name The name in the zip file according to zipfile module.
    #   \return The virtual path of that resource.
    def _zipNameToVirtualPath(self, zip_name: str) -> str:
        if not zip_name.startswith("/"):
            return "/" + zip_name
        return zip_name

    ##  Resize an image to the specified dimensions.
    #
    #   For now you may assume that the input image is PNG formatted.
    #   \param virtual_path The virtual path pointing to an image in the
    #   zipfile.
    #   \param width The desired width of the image.
    #   \param height The desired height of the image.
    #   \return A bytes stream representing a new PNG image with the desired
    #   width and height.
    def _resizeImage(self, virtual_path: str, width: int, height: int) -> IO[bytes]:
        input = self.getStream(virtual_path)
        try:
            from PyQt5.QtGui import QImage
            from PyQt5.QtCore import Qt, QBuffer

            image = QImage()
            image.loadFromData(input.read())
            image = image.scaled(width, height, Qt.IgnoreAspectRatio, Qt.SmoothTransformation)
            output_buffer = QBuffer()
            output_buffer.open(QBuffer.ReadWrite)
            image.save(output_buffer, "PNG")
            output_buffer.seek(0)  # Reset that buffer so that the next guy can request it.
            return BytesIO(output_buffer.readAll())
        except ImportError:
            # TODO: Try other image loaders.
            raise  # Raise import error again if we find no other image loaders.

    #### Below follow some methods to read/write components of the archive. ####

    ##  When loading a file, load the relations from the archive.
    #
    #   If the relations are missing, empty elements are created.
    def _readRels(self) -> None:
        assert self._zipfile is not None

        self._relations[""] = ET.Element("Relationships",
                                         xmlns="http://schemas.openxmlformats.org/package/2006/relationships")  # There must always be a global relationships document.

        # Below is some parsing of paths and extensions.
        # Normally you'd use os.path for this. But this is platform-dependent.
        # For instance, the path separator in Windows is a backslash, but zipfile still uses a slash on Windows.
        # So instead we have custom implementations here. Sorry.

        for virtual_path in self._zipfile.namelist():
            virtual_path = self._zipNameToVirtualPath(virtual_path)
            if not virtual_path.endswith(".rels"):  # We only want to read rels files.
                continue
            directory = virtual_path[:virtual_path.rfind("/")]  # Before the last slash.
            if directory != "_rels" and not directory.endswith("/_rels"):  # Rels files must be in a directory _rels.
                continue

            document = ET.fromstring(self._zipfile.open(virtual_path).read())

            # Find out what file or directory this relation is about.
            origin_filename = virtual_path[virtual_path.rfind("/") + 1:-len(
                ".rels")]  # Just the filename (no path) and without .rels extension.
            origin_directory = directory[
                               :-len("/_rels")]  # The parent path. We already know it's in the _rels directory.
            origin = (origin_directory + "/" if (origin_directory != "") else "") + origin_filename

            self._relations[origin] = document

    ##  At the end of writing a file, write the relations to the archive.
    #
    #   This should be written at the end of writing an archive, when all
    #   relations are known.
    def _writeRels(self) -> None:
        assert self._zipfile is not None
        # Below is some parsing of paths and extensions.
        # Normally you'd use os.path for this. But this is platform-dependent.
        # For instance, the path separator in Windows is a backslash, but zipfile still uses a slash on Windows.
        # So instead we have custom implementations here. Sorry.

        for origin, element in self._relations.items():
            # Find out where to store the rels file.
            if "/" not in origin:  # Is in root.
                origin_directory = ""
                origin_filename = origin
            else:
                origin_directory = origin[:origin.rfind("/")]
                origin_filename = origin[origin.rfind("/") + 1:]
            relations_file = origin_directory + "/_rels/" + origin_filename + ".rels"

            self._indent(element)
            self._zipfile.writestr(relations_file, ET.tostring(self._xml_header) + b"\n" + ET.tostring(element))

    ##  When loading a file, load the content types from the archive.
    #
    #   If the content types are missing, an empty element is created.
    def _readContentTypes(self) -> None:
        assert self._zipfile is not None

        if self._content_types_file in self._zipfile.namelist():
            content_types_element = ET.fromstring(self._zipfile.open(self._content_types_file).read())
            if content_types_element:
                self._content_types_element = content_types_element
        if not self._content_types_element:
            self._content_types_element = ET.Element("Types",
                                                     xmlns="http://schemas.openxmlformats.org/package/2006/content-types")
        # If there is no type for the .rels file, create it.
        if self._mode != OpenMode.ReadOnly:
            for type_element in self._content_types_element.iterfind(
                    "{http://schemas.openxmlformats.org/package/2006/content-types}Default"):
                if "Extension" in type_element.attrib and type_element.attrib["Extension"] == "rels":
                    break
            else:
                ET.SubElement(self._content_types_element, "Default", Extension="rels",
                              ContentType="application/vnd.openxmlformats-package.relationships+xml")

    ##  At the end of writing a file, write the content types to the archive.
    #
    #   This should be written at the end of writing an archive, when all
    #   content types are known.
    def _writeContentTypes(self) -> None:
        assert self._zipfile is not None
        assert self._content_types_element is not None

        self._indent(self._content_types_element)
        self._zipfile.writestr(self._content_types_file,
                               ET.tostring(self._xml_header) + b"\n" + ET.tostring(self._content_types_element))

    ##  When loading a file, read its metadata from the archive.
    #
    #   This depends on the relations! Read the relations first!
    def _readMetadata(self) -> None:
        assert self._zipfile is not None

        for origin, relations_element in self._relations.items():
            for relationship in relations_element.iterfind(
                    "{http://schemas.openxmlformats.org/package/2006/relationships}Relationship"):
                if "Target" not in relationship.attrib or "Type" not in relationship.attrib:  # These two are required, and we actually need them here. Better ignore this one.
                    continue
                if relationship.attrib[
                    "Type"] != self._opc_metadata_relationship_type:  # Not interested in this one. It's not metadata that we recognise.
                    continue
                metadata_file = relationship.attrib["Target"]
                if metadata_file not in self._zipfile.namelist():  # The metadata file is unknown to us.
                    continue

                metadata = json.loads(self._zipfile.open(metadata_file).read().decode("utf-8"))
                if metadata_file == self._global_metadata_file:  # Store globals as if coming from root.
                    metadata_file = ""
                elif metadata_file.endswith(
                        ".json"):  # Metadata files should be named <filename.ext>.json, meaning that they are metadata about <filename.ext>.
                    metadata_file = metadata_file[:-len(".json")]
                self._readMetadataElement(metadata, metadata_file)

        if self._mode != OpenMode.WriteOnly and not self.getMetadata("/3D/model.gcode"):
            try:
                # Check if the G-code file actually exists in the package.
                self._zipfile.getinfo("/3D/model.gcode")
            except KeyError:
                return

            gcode_stream = self._zipfile.open("/3D/model.gcode")
            header_data = GCodeFile.parseHeader(gcode_stream, prefix="/3D/model.gcode/")
            self._metadata.update(header_data)

    ##  Reads a single node of metadata from a JSON document (recursively).
    #   \param element The node in the JSON document to read.
    #   \param current_path The path towards the current document.
    def _readMetadataElement(self, element: Dict[str, Any], current_path: str) -> None:
        for key, value in element.items():
            if isinstance(value, dict):  # json structures stuff in dicts if it is a subtree.
                self._readMetadataElement(value, current_path + "/" + key)
            else:
                self._metadata[current_path + "/" + key] = value

    ##  At the end of writing a file, write the metadata to the archive.
    #
    #   This should be written at the end of writing an archive, when all
    #   metadata is known.
    #
    #   ALWAYS WRITE METADATA BEFORE UPDATING RELS AND CONTENT TYPES.
    def _writeMetadata(self) -> None:
        assert self._zipfile is not None

        keys_left = set(
            self._metadata.keys())  # The keys that are not associated with a particular file (global metadata).
        metadata_per_file = {}  # type: Dict[str, Dict[str, Any]]
        for file_name in self._zipfile.namelist():
            metadata_per_file[file_name] = {}
            for metadata_key in self._metadata:
                if metadata_key.startswith(file_name + "/"):
                    # Strip the prefix: "/a/b/c.stl/print_time" becomes just "print_time" about the file "/a/b/c.stl".
                    metadata_per_file[file_name][metadata_key[len(file_name) + 1:]] = self._metadata[metadata_key]
                    keys_left.remove(metadata_key)
        # keys_left now contains only global metadata keys.

        global_metadata = {key: self._metadata[key] for key in keys_left}
        if len(global_metadata) > 0:
            self._writeMetadataToFile(global_metadata, self._global_metadata_file)
            self.addRelation(self._global_metadata_file, self._opc_metadata_relationship_type)
        for file_name, metadata in metadata_per_file.items():
            if len(metadata) > 0:
                self._writeMetadataToFile(metadata, file_name + ".json")
                self.addRelation(file_name + ".json", self._opc_metadata_relationship_type)
        if len(self._metadata) > 0:  # If we've written any metadata at all, we must include the content type as well.
            try:
                self.addContentType(extension="json", mime_type="text/json")
            except OPCError:  # User may already have defined this content type himself.
                pass

    ##  Writes one dictionary of metadata to a JSON file.
    #   \param metadata The metadata dictionary to write.
    #   \param file_name The virtual path of the JSON file to write to.
    def _writeMetadataToFile(self, metadata: Dict[str, Any], file_name: str) -> None:
        assert self._zipfile is not None

        # Split the metadata into a hierarchical structure.
        document = {}  # type: Dict[str, Any]
        for key, value in metadata.items():
            key = key.strip("/")  # TODO: Should paths ending in a slash give an error?
            path = key.split("/")
            current_element = document
            for element in path:
                if element not in current_element:
                    current_element[element] = {}
                current_element = current_element[element]
            current_element[""] = value

        # We've created some empty-string keys to allow values to occur next to subelements.
        # If this empty-string key is the only key inside a node, fold it in to be just the value.
        for key in metadata:
            key = key.strip("/")
            path = key.split("/")
            current_element = document
            parent = document
            for element in path:
                parent = current_element
                current_element = current_element[element]
            if len(current_element) == 1:  # The empty string is the only element.
                assert "" in current_element
                parent[path[-1]] = current_element[""]  # Fold down the singleton dictionary.

        self._zipfile.writestr(file_name, json.dumps(document, sort_keys=True, indent=4))

    ##  Helper method to write data directly into an aliased path.
    def _writeToAlias(self, path_alias: str, package_filename: str, file_data: bytes) -> None:
        stream = self.getStream("{}/{}".format(path_alias, package_filename))
        stream.write(file_data)

    ##  Helper method to ensure a relationship exists.
    # Creates the relationship if it does not exists, ignores an OPC error if it already does.
    def _ensureRelationExists(self, virtual_path: str, relation_type: str, origin: str) -> None:
        try:
            # We try to add the relation. If this throws an OPCError, we know the relation already exists and ignore it.
            self.addRelation(virtual_path, relation_type, origin)
        except OPCError:
            pass

    ##  Helper function for pretty-printing XML because ETree is stupid.
    #
    #   Source: https://stackoverflow.com/questions/749796/pretty-printing-xml-in-python
    def _indent(self, elem: ET.Element, level: int = 0) -> None:
        i = "\n" + level * "  "
        if len(elem):
            if not elem.text or not elem.text.strip():
                elem.text = i + "  "
            if not elem.tail or not elem.tail.strip():
                elem.tail = i
            for elem in elem:
                self._indent(elem, level + 1)
            if not elem.tail or not elem.tail.strip():
                elem.tail = i
        else:
            if level and (not elem.tail or not elem.tail.strip()):
                elem.tail = i


##  Error to raise that something went wrong with reading/writing a OPC file.
class OPCError(Exception):
    pass  # This is just a marker class.




import os
import logging

import dbus.service
import dbus.mainloop.glib
from typing import Dict, Any
from gi.repository import GLib

import Charon.Service

# Very basic service main loop built with GLib.

GLib.threads_init()
dbus.mainloop.glib.threads_init()

config = {} # type: Dict[str, Any]
if os.environ.get("CHARON_DEBUG", "0") == "1":
    config["level"] = logging.DEBUG
else:
    config["level"] = logging.WARNING
config["format"] = "%(asctime)s | %(levelname)s | %(name)s:%(lineno)d@%(funcName)s | %(message)s"
logging.basicConfig(**config)

_loop = GLib.MainLoop()

# Use a single bus object for all dbus communication.
if os.environ.get("CHARON_USE_SESSION_BUS", "1") == "1":
    _bus = dbus.SessionBus(private=True, mainloop=dbus.mainloop.glib.DBusGMainLoop())
else:
    _bus = dbus.SystemBus(private=True, mainloop=dbus.mainloop.glib.DBusGMainLoop())

_service = Charon.Service.FileService(_bus)

_loop.run()



# Copyright (c) 2018 Ultimaker B.V.
# libCharon is released under the terms of the LGPLv3 or higher.
import ast

from typing import Any, Dict, IO, List, Optional, Union

from Charon.FileInterface import FileInterface
from Charon.OpenMode import OpenMode


def isAPositiveNumber(a: str) -> bool:
    try:
        number = float(repr(a))
        return number >= 0
    except:
        bool_a = False

    return bool_a


class GCodeFile(FileInterface):
    mime_type = "text/x-gcode"

    MaximumHeaderLength = 100

    def __init__(self) -> None:
        self.__stream = None  # type: Optional[IO[bytes]]
        self.__metadata = {}  # type: Dict[str, Any]

    def openStream(self, stream: IO[bytes], mime: str, mode: OpenMode = OpenMode.ReadOnly) -> None:
        if mode != OpenMode.ReadOnly:
            raise NotImplementedError()

        self.__stream = stream
        self.__metadata = {}
        self.__metadata = self.parseHeader(self.__stream, prefix = "/metadata/toolpath/default/")

    @staticmethod
    def parseHeader(stream: IO[bytes], *, prefix: str = "") -> Dict[str, Any]:
        try:
            metadata = {} # type: Dict[str, Any]
            line_number = 0
            for line_number, bytes_line in enumerate(stream):
                if line_number > GCodeFile.MaximumHeaderLength:
                    break
                line = bytes_line.decode("utf-8")

                if line.startswith(";START_OF_HEADER"):
                    continue
                elif line.startswith(";LAYER") or line.startswith(";END_OF_HEADER"):
                    break
                elif line.startswith(";HEADER_VERSION"):
                    # Header version is a number but should not be parsed as number, so special case it.
                    metadata["header_version"] = line.split(":")[1].strip()
                elif line.startswith(";") and ":" in line:
                    key, value = line[1:].split(":")
                    key = key.strip().lower()
                    value = value.strip()
                    try:
                        value = ast.literal_eval(value.strip())
                    except:
                        pass
                    key_elements = key.split(".")
                    GCodeFile.__insertKeyValuePair(metadata, key_elements, value)
            
            if stream.seekable():
                stream.seek(0)

            flavor = metadata.get("flavor", None)
            if flavor == "Griffin":
                if metadata["header_version"] != "0.1":
                    raise InvalidHeaderException("Unsupported Griffin header version: {0}".format(metadata["header_version"]))
                GCodeFile.__validateGriffinHeader(metadata)
                GCodeFile.__cleanGriffinHeader(metadata)
            elif flavor == "UltiGCode":
                metadata["machine_type"] = "ultimaker2"
            else:
                raise InvalidHeaderException("Flavor must be defined!")

            if prefix:
                prefixed_metadata = {}
                for key, value in metadata.items():
                    prefixed_metadata[prefix + key] = value
                metadata = prefixed_metadata

            return metadata
        except Exception as e:
            raise InvalidHeaderException("Unable to parse the header. An exception occured; %s" % e)


    ## Add a key-value pair to the metadata dictionary.
    # Splits up key each element to it's own dictionary.
    # @param metadata Metadata collection
    # @param key_elements List of separate key name elements
    # @param value Key value
    @staticmethod
    def __insertKeyValuePair(
        metadata: Dict[str, Any],
        key_elements: Any,
        value: Any
    ) -> Any:
        if not key_elements:
            return value 

        sub_dict = {}

        if key_elements[0] in metadata:
            sub_dict = metadata[key_elements[0]]

        metadata[key_elements[0]] = GCodeFile.__insertKeyValuePair(sub_dict, key_elements[1:], value)
       
        return metadata    

    def getData(self, virtual_path: str) -> Dict[str, Any]:
        assert self.__stream is not None

        if virtual_path.startswith("/metadata"):
            result = {}
            for key, value in self.__metadata.items():
                if key.startswith(virtual_path):
                    result[key] = value
            return result

        if virtual_path == "/toolpath" or virtual_path == "/toolpath/default":
            return { virtual_path: self.__stream.read() }

        return {}

    ## Cleans a parsed GRIFFIN flavoured GCODE header.
    @staticmethod
    def __cleanGriffinHeader(metadata: Dict[str, Any]) -> None:
        metadata["machine_type"] = metadata["target_machine"]["name"]
        del metadata["target_machine"]
    
        if GCodeFile.__isAvailable(metadata, ["time"]):
            GCodeFile.__insertKeyValuePair(metadata, ["print", "time"], metadata["time"])
            # del metadata["time"]  # We want to delete the old key, but it's behavior of how the code was.

        GCodeFile.__insertKeyValuePair(metadata, ["print", "min_size"], metadata["print"]["size"]["min"])
        GCodeFile.__insertKeyValuePair(metadata, ["print", "max_size"], metadata["print"]["size"]["max"])
        del metadata["print"]["size"]
        
        for key, value in metadata["extruder_train"].items():
            GCodeFile.__insertKeyValuePair(metadata, ["extruders", int(key)], value)

        del metadata["extruder_train"]

    ## Checks if a path to a key is available
    # @param metadata Metadata collection to check for the presence of the key
    # @param keys List of key elements describing the path to a value. If a key element is a list, then all the elements
    #             must exist on the location of that key element
    # @return True if the key is available and not empty
    @staticmethod
    def __isAvailable(metadata: Dict[str, Any], keys: List[Any]) -> bool:
        if not keys:
            return True

        key = keys[0]
        
        if isinstance(key, list):
            key_is_valid = True
            for sub_key in key:
                key_is_valid = key_is_valid and GCodeFile.__isAvailable(metadata, [sub_key] + [keys[1:]])
        else:
            key_is_valid = key in metadata and metadata[key] is not None and not str(metadata[key]) == ""
            key_is_valid = key_is_valid and GCodeFile.__isAvailable(metadata[key], keys[1:])

        return key_is_valid
    
    ## Validates a parsed GRIFFIN flavoured GCODE header.
    # Will raise an InvalidHeader exception when the header is invalid.
    # @param metadata Key/value dictionary based on the header.
    @staticmethod
    def __validateGriffinHeader(metadata: Dict[str, Any]) -> None:

        # Validate target settings
        if not GCodeFile.__isAvailable(metadata, ["target_machine", "name"]): 
            raise InvalidHeaderException("TARGET_MACHINE.NAME must be set")

        # Validate generator settings
        if not GCodeFile.__isAvailable(metadata, ["generator", "name"]): 
            raise InvalidHeaderException("GENERATOR.NAME must be set")
        if not GCodeFile.__isAvailable(metadata, ["generator", "version"]): 
            raise InvalidHeaderException("GENERATOR.VERSION must be set")
        if not GCodeFile.__isAvailable(metadata, ["generator", "build_date"]): 
            raise InvalidHeaderException("GENERATOR.BUILD_DATE must be set")

        # Validate build plate temperature 
        if not GCodeFile.__isAvailable(metadata, ["build_plate", "initial_temperature"]) or \
            not isAPositiveNumber(metadata["build_plate"]["initial_temperature"]):
            raise InvalidHeaderException("BUILD_PLATE.INITIAL_TEMPERATURE must be set and be a positive real")
      
        # Validate dimensions 
        if not GCodeFile.__isAvailable(metadata, ["print", "size", "min", ["x", "y", "z"]]):
            raise InvalidHeaderException("PRINT.SIZE.MIN.[x,y,z] must be set. Ensure all three are defined.")
        if not GCodeFile.__isAvailable(metadata, ["print", "size", "max", ["x", "y", "z"]]):
            raise InvalidHeaderException("PRINT.SIZE.MAX.[x,y,z] must be set. Ensure all three are defined.")
        
        # Validate print time
        print_time = -1

        if GCodeFile.__isAvailable(metadata, ["print", "time"]):
            print_time = int(metadata["print"]["time"])
        elif GCodeFile.__isAvailable(metadata, ["time"]):
            print_time = int(metadata["time"])
        else:
            raise InvalidHeaderException("TIME or PRINT.TIME must be set")

        if print_time < 0:
            raise InvalidHeaderException("Print Time should be a positive integer")
       
        # Validate extruder train
        for index in range(0, 10):
            index_str = str(index)
            if GCodeFile.__isAvailable(metadata, ["extruder_train", index_str]):

                if not GCodeFile.__isAvailable(metadata, ["extruder_train", index_str, "nozzle", "diameter"]) or \
                    not isAPositiveNumber(metadata["extruder_train"][index_str]["nozzle"]["diameter"]):
                        raise InvalidHeaderException(
                            "extruder_train.{}.nozzle.diameter must be defined and be a positive real".format(index))
       
                if not GCodeFile.__isAvailable(metadata, ["extruder_train", index_str, "material", "volume_used"]) or \
                    not isAPositiveNumber(metadata["extruder_train"][index_str]["material"]["volume_used"]):
                        raise InvalidHeaderException(
                            "extruder_train.{}.material.volume_used must be defined and positive".format(index))

                if not GCodeFile.__isAvailable(metadata, ["extruder_train", index_str, "initial_temperature"]) or \
                    not isAPositiveNumber(metadata["extruder_train"][index_str]["initial_temperature"]):
                        raise InvalidHeaderException(
                            "extruder_train.{}.initial_temperature must be defined and positive".format(index))

    def getStream(self, virtual_path: str) -> IO[bytes]:
        assert self.__stream is not None
        
        if virtual_path != "/toolpath" and virtual_path != "/toolpath/default":
            raise NotImplementedError("G-code files only support /toolpath as stream")

        return self.__stream

    def close(self) -> None:
        assert self.__stream is not None
        
        self.__stream.close()


class InvalidHeaderException(Exception):
    pass



# Copyright (c) 2018 Ultimaker B.V.
# libCharon is released under the terms of the LGPLv3 or higher.
import gzip

from Charon.filetypes.GCodeFile import GCodeFile


class GCodeGzFile(GCodeFile):
    stream_handler = gzip.open
    mime_type = "text/x-gcode-gz"

    def __init__(self) -> None:
        super().__init__()



# Copyright (c) 2018 Ultimaker B.V.
# libCharon is released under the terms of the LGPLv3 or higher.

import enum #The class in this file is an enum.

##  The possible purposes for which you could open a file.
#
#   You could always open a file in read-write mode, but it's best practice to
#   open a file in specific read or write only modes if only one of the two is
#   needed. This will prevent the programmer from accidentally modifying the
#   file and may trigger some operating systems to treat the file lock
#   differently.
class OpenMode(enum.Enum):
    ##  The file can only be read from.
    ReadOnly = "r"

    ##  The file can only be written to.
    WriteOnly = "w"


# Copyright (c) 2018 Ultimaker B.V.
# libCharon is released under the terms of the LGPLv3 or higher.
from typing import Any, Dict, List, IO, Optional, Callable

from Charon.OpenMode import OpenMode


##  An interface for accessing files.
#
#   This interface is designed to be able to access 3D-printing related files,
#   and for container-type files to access the resources therein.
class FileInterface:
    stream_handler = open # type: Callable[[str, str], IO[bytes]]

    mime_type = ""

    ##  Opens a file for reading or writing.
    #
    #   After opening the file, this instance will represent that file from then
    #   on, meaning that the metadata getters/setters and the streams will be
    #   functioning on that file.
    #   \param path The path to the file on local disk, relative or absolute.
    #   \param mode The mode with which to open the file (see OpenMode).
    def open(self, path: str, mode: OpenMode = OpenMode.ReadOnly) -> None:
        raise NotImplementedError("The open() function of " + self.__class__.__qualname__ + " is not implemented.")

    ##  Opens a stream for reading or writing.
    #
    #   After opening the stream, this instance will represent that stream from
    #   then on, meaning that the metadata getters/setters and the streams will
    #   be functioning on that stream.
    #   \param stream The stream to read from or write to.
    #   \param mime The MIME type of the stream. This determines what implementation is used to read/write it.
    #   \param mode The mode with which to open the file (see OpenMode).
    def openStream(self, stream: IO[bytes], mime: str, mode: OpenMode = OpenMode.ReadOnly) -> None:
        raise NotImplementedError("The openStream() function of " + self.__class__.__qualname__ + " is not implemented.")

    ##  Closes the opened file, releasing the resources in use for it.
    #
    #   After the file is closed, this instance can no longer be used until the ``open`` method is called again.
    def close(self) -> None:
        raise NotImplementedError("The close() function of " + self.__class__.__qualname__ + " is not implemented.")

    ##  Ensures that no buffered data is still pending to be read or written.
    def flush(self) -> None:
        raise NotImplementedError("The flush() function of " + self.__class__.__qualname__ + " is not implemented.")

    ##  Returns a list of all resources and metadata in the file.
    def listPaths(self) -> List[str]:
        raise NotImplementedError("The listPaths() function of " + self.__class__.__qualname__ + " is not implemented.")

    ##  Gets the data stored at the specified virtual path and all its descendants.
    #
    #   The returned dictionary may contain normal resources as well as
    #   metadata. If it is a normal resource, the value will contain the
    #   serialised data (either ``bytes`` or ``str``, depending on whether the
    #   file opens in binary mode or not). If it is metadata, all metadata keys
    #   under the specified path are returned (all descendants in the tree). If
    #   there is no metadata and no resource under the selected virtual path, an
    #   empty dictionary is returned.
    #   \param virtual_path The path inside the file to get the data from.
    #   \return The data and metadata under the specified virtual path.
    def getData(self, virtual_path: str) -> Dict[str, Any]:
        raise NotImplementedError("The getData() function of " + self.__class__.__qualname__ + " is not implemented.")

    ##  Sets the data of several virtual paths at once.
    #
    #   The ``data`` parameter provides a dictionary mapping virtual paths to
    #   the new data that should be provided in the path.
    def setData(self, data: Dict[str, Any]) -> None:
        raise NotImplementedError("The setData() function of " + self.__class__.__qualname__ + " is not implemented.")

    ##  Gets metadata entries in the opened file.
    #
    #   The metadata is a dictionary, where the keys are virtual paths in the
    #   subtree of the resource tree specified by ``virtual_path``. For
    #   instance, when requesting the metadata of the resource with virtual path
    #   ``/metadata``, this function could return a dictionary containing:
    #   * ``/metadata/size``: 12354
    #   * ``/metadata/toolpath/default/size``: 12000
    #   * ``/metadata/toolpath/default/machine_type``: ``ultimaker3``
    #   * ``/metadata/toolpath/default/print_time``: 121245
    #   * ``/metadata/toolpath/default/print_size``: (0, 0, 0) x (100, 100, 100)
    #
    #   But a subtree can be requested as well, such as
    #   ``/metadata/toolpath/default/size``, which would then return a
    #   dictionary containing only the key ``/metadata/toolpath/default/size``
    #   and its value, because there are no other subitems in that subtree.
    #
    #   If there is no metadata in the requested path, an empty dictionary is
    #   returned.
    #   \param virtual_path The subtree of metadata entries to get the metadata
    #   of.
    #   \return A dictionary of all the metadata entries in the selected
    #   subtree.
    def getMetadata(self, virtual_path: str) -> Dict[str, Any]:
        raise NotImplementedError("The getMetadata() function of " + self.__class__.__qualname__ + " is not implemented.")

    ##  Changes some metadata entries in the opened file.
    #
    #   The provided dictionary must have the full virtual paths of the metadata
    #   entries it wants to change as its keys, and the new values along with
    #   every key.
    #
    #   If a metadata entry didn't exist yet, it is created.
    #
    #   If a metadata entry by cannot be changed (such as the file size of a
    #   resource) then a ``ReadOnlyError`` must be raised for that resource, and
    #   none of the changes of this function call may be applied (or everything
    #   must be undone).
    #   \param metadata A dictionary of metadata entries to change.
    #   \raises ReadOnlyError A metadata entry cannot be changed (such as the
    #   file size of a resource).
    def setMetadata(self, metadata: Dict[str, Any]) -> None:
        raise NotImplementedError("The setMetadata() function of " + self.__class__.__qualname__ + " is not implemented.")

    ##  Gets an I/O stream to the resource or metadata at the specified virtual
    #   path.
    #
    #   This stream may be a normal resource or it may be metadata. If it is
    #   metadata, a stream will be returned in the form of a JSON document
    #   (encoded in UTF-8 for binary streams) containing all the metadata that
    #   would be returned by the getMetadata method.
    #
    #   Whether the returned stream is an input or an output stream depends on
    #   the mode that was provided in the ``open`` method. This determines
    #   whether you can read from and/or write to the stream.
    #
    #   If a resource didn't exist and you can write, the resource is created.
    #   \param virtual_path The virtual path to the resource that you want to
    #   read or write.
    #   \raises ReadOnlyError The resource doesn't exist and there are no write
    #   permissions to create it.
    def getStream(self, virtual_path: str) -> IO[bytes]:
        raise NotImplementedError("The getStream() function of " + self.__class__.__qualname__ + " is not implemented.")

    ##  Gets a bytes representation of the file.
    #
    #   Resources inside the file are not supported by this method. Use
    #   ``getStream`` for that.
    #   \param offset The number of bytes to skip at the beginning of the file.
    #   \param count The maximum number of bytes to return. If the file is
    #   longer than this, it is truncated. If the file is shorter than this,
    #   fewer bytes than this might be returned. If not specified, the entire
    #   file will be returned except the initial offset.
    #   \return bytes A bytes array representing the file or a part of it.
    def toByteArray(self, offset: int = 0, count: int = -1) -> bytes:
        raise NotImplementedError("The toByteArray() function of " + self.__class__.__qualname__ + " is not implemented.")


import io
import unittest

from Charon.filetypes.GCodeFile import GCodeFile, InvalidHeaderException


class TestGcodeFile(unittest.TestCase):

    __minimal_griffin_header = ";START_OF_HEADER\n" \
        ";FLAVOR:Griffin\n" \
        ";TARGET_MACHINE.NAME:target.foobar\n" \
        ";GENERATOR.NAME:generator_foo\n" \
        ";GENERATOR.VERSION: generator_version_foo\n" \
        ";GENERATOR.BUILD_DATE: generator_build_foo\n" \
        ";BUILD_PLATE.INITIAL_TEMPERATURE:30\n" \
        ";PRINT.SIZE.MIN.X:1\n" \
        ";PRINT.SIZE.MIN.Y:2\n" \
        ";PRINT.SIZE.MIN.Z:3\n" \
        ";PRINT.SIZE.MAX.X:1\n" \
        ";PRINT.SIZE.MAX.Y:2\n" \
        ";PRINT.SIZE.MAX.Z:3\n" \
        ";HEADER_VERSION:0.1\n" \
        ";TIME:11\n" \
        ";EXTRUDER_TRAIN.1.NOZZLE.DIAMETER:1.5\n" \
        ";EXTRUDER_TRAIN.1.MATERIAL.VOLUME_USED:1.5\n" \
        ";EXTRUDER_TRAIN.1.INITIAL_TEMPERATURE:666\n" \
        "{}\n" \
        ";END_OF_HEADER"

    def _print(self, d, prefix=""):
        for k, v in d.items():
            if type(v) is dict:
                self._print(v, prefix="{}.{}".format(prefix, k) if prefix else "{}".format(k))
            else:
                if prefix:
                    print("{}.{}: {}".format(prefix, k, v))
                else:
                    print("{}: {}".format(k, v))

    def testParseGenericParameter_HappyTrail(self) -> None:
        gcode = self.__minimal_griffin_header.format(";A.B.C:5")
        gcode_stream = io.BytesIO(str.encode(gcode))
        metadata = GCodeFile.parseHeader(gcode_stream)

        self._print(metadata)  # print if any assert fails
        self.assertEqual(metadata["a"]["b"]["c"], 5)
        self.assertEqual(metadata["generator"]["name"], "generator_foo")
        self.assertEqual(metadata["build_plate"]["initial_temperature"], 30)
        self.assertEqual(metadata["extruders"][1]["nozzle"]["diameter"], 1.5)
        self.assertEqual(metadata["print"]["time"], 11)
        self.assertEqual(metadata["time"], 11)  # This was the behavior of the old code.

    def testParseHeader_MissingFlavor(self) -> None:
        gcode = ";START_OF_HEADER\n" \
            ";TARGET_MACHINE.NAME:target.foobar\n" \
            ";GENERATOR.NAME:generator_foo\n" \
            ";GENERATOR.VERSION: generator_version_foo\n" \
            ";GENERATOR.BUILD_DATE: generator_build_foo\n" \
            ";BUILD_PLATE.INITIAL_TEMPERATURE:30\n" \
            ";PRINT.SIZE.MIN.X:1\n" \
            ";PRINT.SIZE.MIN.Y:2\n" \
            ";PRINT.SIZE.MIN.Z:3\n" \
            ";PRINT.SIZE.MAX.X:1\n" \
            ";PRINT.SIZE.MAX.Y:2\n" \
            ";PRINT.SIZE.MAX.Z:3\n" \
            ";HEADER_VERSION:0.1\n" \
            ";TIME:11\n"

        self.__parseWithInvalidHeaderException(gcode, "Flavor")

    def testParseHeader_MissingHeaderVersion(self) -> None:
        gcode = ";START_OF_HEADER\n" \
            ";FLAVOR:Griffin\n" \
            ";TARGET_MACHINE.NAME:target.foobar\n" \
            ";GENERATOR.NAME:generator_foo\n" \
            ";GENERATOR.VERSION: generator_version_foo\n" \
            ";GENERATOR.BUILD_DATE: generator_build_foo\n" \
            ";BUILD_PLATE.INITIAL_TEMPERATURE:30\n" \
            ";PRINT.SIZE.MIN.X:1\n" \
            ";PRINT.SIZE.MIN.Y:2\n" \
            ";PRINT.SIZE.MIN.Z:3\n" \
            ";PRINT.SIZE.MAX.X:1\n" \
            ";PRINT.SIZE.MAX.Y:2\n" \
            ";PRINT.SIZE.MAX.Z:3\n" \
            ";TIME:11\n"

        self.__parseWithInvalidHeaderException(gcode, "version")

    def testParseHeader_MissingTargetMachine(self) -> None:
        gcode = ";START_OF_HEADER\n" \
            ";FLAVOR:Griffin\n" \
            ";GENERATOR.NAME:generator_foo\n" \
            ";GENERATOR.VERSION: generator_version_foo\n" \
            ";GENERATOR.BUILD_DATE: generator_build_foo\n" \
            ";BUILD_PLATE.INITIAL_TEMPERATURE:30\n" \
            ";PRINT.SIZE.MIN.X:1\n" \
            ";PRINT.SIZE.MIN.Y:2\n" \
            ";PRINT.SIZE.MIN.Z:3\n" \
            ";PRINT.SIZE.MAX.X:1\n" \
            ";PRINT.SIZE.MAX.Y:2\n" \
            ";PRINT.SIZE.MAX.Z:3\n" \
            ";HEADER_VERSION:0.1\n" \
            ";TIME:11\n"
        self.__parseWithInvalidHeaderException(gcode, "TARGET_MACHINE")

    def testParseHeader_MissingGeneratorName(self) -> None:
        gcode = ";START_OF_HEADER\n" \
            ";FLAVOR:Griffin\n" \
            ";TARGET_MACHINE.NAME:target.foobar\n" \
            ";GENERATOR.VERSION: generator_version_foo\n" \
            ";GENERATOR.BUILD_DATE: generator_build_foo\n" \
            ";BUILD_PLATE.INITIAL_TEMPERATURE:30\n" \
            ";PRINT.SIZE.MIN.X:1\n" \
            ";PRINT.SIZE.MIN.Y:2\n" \
            ";PRINT.SIZE.MIN.Z:3\n" \
            ";PRINT.SIZE.MAX.X:1\n" \
            ";PRINT.SIZE.MAX.Y:2\n" \
            ";PRINT.SIZE.MAX.Z:3\n" \
            ";HEADER_VERSION:0.1\n" \
            ";TIME:11\n"
        self.__parseWithInvalidHeaderException(gcode, "GENERATOR.NAME")

    def testParseHeader_MissingGeneratorVersion(self) -> None:
        gcode = ";START_OF_HEADER\n" \
            ";FLAVOR:Griffin\n" \
            ";TARGET_MACHINE.NAME:target.foobar\n" \
            ";GENERATOR.NAME:generator_foo\n" \
            ";GENERATOR.BUILD_DATE: generator_build_foo\n" \
            ";BUILD_PLATE.INITIAL_TEMPERATURE:30\n" \
            ";PRINT.SIZE.MIN.X:1\n" \
            ";PRINT.SIZE.MIN.Y:2\n" \
            ";PRINT.SIZE.MIN.Z:3\n" \
            ";PRINT.SIZE.MAX.X:1\n" \
            ";PRINT.SIZE.MAX.Y:2\n" \
            ";PRINT.SIZE.MAX.Z:3\n" \
            ";HEADER_VERSION:0.1\n" \
            ";TIME:11\n"
        self.__parseWithInvalidHeaderException(gcode, "GENERATOR.VERSION")

    def testParseHeader_MissingGeneratorBuildDate(self) -> None:
        gcode = ";START_OF_HEADER\n" \
            ";FLAVOR:Griffin\n" \
            ";TARGET_MACHINE.NAME:target.foobar\n" \
            ";GENERATOR.NAME:generator_foo\n" \
            ";GENERATOR.VERSION: generator_version_foo\n" \
            ";BUILD_PLATE.INITIAL_TEMPERATURE:30\n" \
            ";PRINT.SIZE.MIN.X:1\n" \
            ";PRINT.SIZE.MIN.Y:2\n" \
            ";PRINT.SIZE.MIN.Z:3\n" \
            ";PRINT.SIZE.MAX.X:1\n" \
            ";PRINT.SIZE.MAX.Y:2\n" \
            ";PRINT.SIZE.MAX.Z:3\n" \
            ";HEADER_VERSION:0.1\n" \
            ";TIME:11\n"
        self.__parseWithInvalidHeaderException(gcode, "GENERATOR.BUILD_DATE")

    def testParseHeader_MissingInitialBuildPlateTemp(self) -> None:
        gcode = ";START_OF_HEADER\n" \
            ";FLAVOR:Griffin\n" \
            ";TARGET_MACHINE.NAME:target.foobar\n" \
            ";GENERATOR.NAME:generator_foo\n" \
            ";GENERATOR.VERSION: generator_version_foo\n" \
            ";GENERATOR.BUILD_DATE: generator_build_foo\n" \
            ";PRINT.SIZE.MIN.X:1\n" \
            ";PRINT.SIZE.MIN.Y:2\n" \
            ";PRINT.SIZE.MIN.Z:3\n" \
            ";PRINT.SIZE.MAX.X:1\n" \
            ";PRINT.SIZE.MAX.Y:2\n" \
            ";PRINT.SIZE.MAX.Z:3\n" \
            ";HEADER_VERSION:0.1\n" \
            ";TIME:11\n"

        self.__parseWithInvalidHeaderException(gcode, "BUILD_PLATE.INITIAL_TEMPERATURE")

    def testParseHeader_MissingMinSizeX(self) -> None:
        gcode = ";START_OF_HEADER\n" \
            ";FLAVOR:Griffin\n" \
            ";TARGET_MACHINE.NAME:target.foobar\n" \
            ";GENERATOR.NAME:generator_foo\n" \
            ";GENERATOR.VERSION: generator_version_foo\n" \
            ";GENERATOR.BUILD_DATE: generator_build_foo\n" \
            ";BUILD_PLATE.INITIAL_TEMPERATURE:30\n" \
            ";PRINT.SIZE.MIN.Y:2\n" \
            ";PRINT.SIZE.MIN.Z:3\n" \
            ";PRINT.SIZE.MAX.X:1\n" \
            ";PRINT.SIZE.MAX.Y:2\n" \
            ";PRINT.SIZE.MAX.Z:3\n" \
            ";HEADER_VERSION:0.1\n" \
            ";TIME:11\n"

        self.__parseWithInvalidHeaderException(gcode, "PRINT.SIZE.MIN")

    def testParseHeader_MissingMaxSizeZ(self) -> None:
        gcode = ";START_OF_HEADER\n" \
            ";FLAVOR:Griffin\n" \
            ";TARGET_MACHINE.NAME:target.foobar\n" \
            ";GENERATOR.NAME:generator_foo\n" \
            ";GENERATOR.VERSION: generator_version_foo\n" \
            ";GENERATOR.BUILD_DATE: generator_build_foo\n" \
            ";BUILD_PLATE.INITIAL_TEMPERATURE:30\n" \
            ";PRINT.SIZE.MIN.X:1\n" \
            ";PRINT.SIZE.MIN.Y:2\n" \
            ";PRINT.SIZE.MIN.Z:3\n" \
            ";PRINT.SIZE.MAX.X:1\n" \
            ";PRINT.SIZE.MAX.Y:2\n" \
            ";HEADER_VERSION:0.1\n" \
            ";TIME:11\n"

        self.__parseWithInvalidHeaderException(gcode, "PRINT.SIZE.MAX")

    def testParseHeader_MissingPrintTime(self) -> None:
        gcode = ";START_OF_HEADER\n" \
            ";FLAVOR:Griffin\n" \
            ";TARGET_MACHINE.NAME:target.foobar\n" \
            ";GENERATOR.NAME:generator_foo\n" \
            ";GENERATOR.VERSION: generator_version_foo\n" \
            ";GENERATOR.BUILD_DATE: generator_build_foo\n" \
            ";BUILD_PLATE.INITIAL_TEMPERATURE:30\n" \
            ";PRINT.SIZE.MIN.X:1\n" \
            ";PRINT.SIZE.MIN.Y:2\n" \
            ";PRINT.SIZE.MIN.Z:3\n" \
            ";PRINT.SIZE.MAX.X:1\n" \
            ";PRINT.SIZE.MAX.Y:2\n" \
            ";PRINT.SIZE.MAX.Z:3\n" \
            ";HEADER_VERSION:0.1\n"

        self.__parseWithInvalidHeaderException(gcode, "TIME")
        self.__parseWithInvalidHeaderException(gcode, "PRINT.TIME")

    def __parseWithInvalidHeaderException(self, gcode, text) -> None: 
        gcode_stream = io.BytesIO(str.encode(gcode))
        
        with self.assertRaises(InvalidHeaderException) as cm:
            metadata = GCodeFile.parseHeader(gcode_stream)
        self.assertTrue(text  in str(cm.exception))

    def testParseGenericParameter_NoValue(self) -> None:
        gcode = self.__minimal_griffin_header.format(";A.B.C:")
        gcode_stream = io.BytesIO(str.encode(gcode))
        metadata = GCodeFile.parseHeader(gcode_stream)

        self.assertEqual(metadata["a"]["b"]["c"], '')




# Copyright (c) 2018 Ultimaker B.V.
# libCharon is released under the terms of the LGPLv3 or higher.
from collections import OrderedDict

from Charon.OpenMode import OpenMode
from Charon.filetypes.GCodeFile import GCodeFile
from Charon.filetypes.OpenPackagingConvention import OpenPackagingConvention


##  A container file type that contains multiple 3D-printing related files that belong together.
class UltimakerFormatPackage(OpenPackagingConvention):

    # Where the global metadata file is.
    _global_metadata_file = "/Metadata/UFP_Global.json"

    # Unique identifier of the relationship type that relates UFP metadata to files.
    _metadata_relationship_type = "http://schemas.ultimaker.org/package/2018/relationships/ufp_metadata"

    # Where the global metadata file is.
    global_metadata_file = "/Metadata/UFP_Global.json"

    # Unique identifier of the relationship type that relates UFP metadata to files.
    metadata_relationship_type = "http://schemas.ultimaker.org/package/2018/relationships/ufp_metadata"

    # Virtual path aliases. Keys are regex. Order matters!
    _aliases = OrderedDict([
        (r"^/preview/default", "/Metadata/thumbnail.png"),
        (r"^/preview", "/Metadata/thumbnail.png"),
        (r"^/toolpath/default", "/3D/model.gcode"),
        (r"^/toolpath", "/3D/model.gcode"),
    ])

    mime_type = "application/x-ufp"

    ##  Initialises the fields of this class.
    def __init__(self):
        super().__init__()

    ##  When loading a file, read its metadata from the archive.
    #
    #   This depends on the relations! Read the relations first!
    def _readMetadata(self) -> None:
        super()._readMetadata()
        if self._mode != OpenMode.WriteOnly and not self.getMetadata("/3D/model.gcode"):
            try:
                # Check if the G-code file actually exists in the package.
                self._zipfile.getinfo("/3D/model.gcode")
            except KeyError:
                return

            gcode_stream = self._zipfile.open("/3D/model.gcode")
            header_data = GCodeFile.parseHeader(gcode_stream, prefix="/3D/model.gcode/")
            self._metadata.update(header_data)



import sys

from PyQt5.QtCore import QCoreApplication, QTimer

import Charon.Client

if len(sys.argv) != 2:
    print("Usage: test.py [file]")
    exit(1)

app = QCoreApplication(sys.argv)

request = Charon.Client.Request(sys.argv[1], ["/Metadata/thumbnail.png"])
request.start()

while(request.state == Charon.Client.Request.State.Running):
    app.processEvents()

if request.state == Charon.Client.Request.State.Completed:
    print("Request Complete")
    print(request.data)
elif request.state == Charon.Client.Request.State.Error:
    print("Request Error")
    print(request.errorString)
else:
    print("Request did not finish properly")
    print(request.state)



# Copyright (c) 2018 Ultimaker B.V.
# Charon is released under the terms of the LGPLv3 or higher.
import os

from Charon.VirtualFile import VirtualFile


def test_GCodeReader():
    f = VirtualFile()
    f.open(os.path.join(os.path.dirname(__file__), "resources", "um3.gcode"))
    assert f.getData("/metadata")["/metadata/toolpath/default/flavor"] == "Griffin"
    assert b"M104" in f.getStream("/toolpath").read()
    f.close()


def test_GCodeGzReader():
    f = VirtualFile()
    f.open(os.path.join(os.path.dirname(__file__), "resources", "um3.gcode.gz"))
    assert f.getData("/metadata")["/metadata/toolpath/default/flavor"] == "Griffin"
    assert b"M104" in f.getStream("/toolpath").read()
    f.close()



import dbus
import logging

import RequestQueue

log = logging.getLogger(__name__)

##  The main interface for the Charon file service.
#
#   This contains the main interface definition for the Charon file service.
#   It is exposed over DBus as the "nl.ultimaker.charon" service, with
#   "/nl/ultimaker/charon" as its object path and all functions registered
#   in the "nl.ultimaker.charon" interface name.
#
#   The file service maintains a queue of jobs that need to be processed.
#   See RequestQueue for details on this process.
#
#   Note: This class does not currently use type hinting since type hints,
#   dbus-python decorators and Python 3.4 do not mix well.
class FileService(dbus.service.Object):
    def __init__(self, dbus_bus: dbus.Bus) -> None:
        super().__init__(
            bus_name = dbus.service.BusName("nl.ultimaker.charon", dbus_bus),
            object_path = "/nl/ultimaker/charon"
        )

        log.debug("FileService initialized")
        self.__queue = RequestQueue.RequestQueue()

    ##  Start a request for data from a file.
    #
    #   This function will start a request for data from a certain file.
    #   It will be processed in a separate thread.
    #
    #   When the request has finished, `requestFinished` will be emitted.
    #
    #   \param request_id A unique identifier to track this request with.
    #   \param file_path The path to a file to load.
    #   \param virtual_paths A list of virtual paths that define what set of data to retrieve.
    #
    #   \return A boolean indicating whether the request was successfully started.
    @dbus.decorators.method("nl.ultimaker.charon", "ssas", "b")
    def startRequest(self, request_id, file_path, virtual_paths):
        log.debug("Received request {id} for {virtual} from {path}".format(id = request_id, virtual = virtual_paths, path = file_path))
        request = RequestQueue.Request(self, request_id, file_path, virtual_paths)
        return self.__queue.enqueue(request)

    ##  Cancel a pending request for data.
    #
    #   This will cancel a request that was previously posted.
    #
    #   Note that if the request is already being processed, the request will not be
    #   canceled. If the cancel was successful, `requestError` will be emitted with the
    #   specified request and an error string describing it was canceled.
    #
    #   \param request_id The ID of the request to cancel.
    @dbus.decorators.method("nl.ultimaker.charon", "s", "")
    def cancelRequest(self, request_id):
        log.debug("Cancel request '{id}'".format(id = request_id))
        if self.__queue.dequeue(request_id):
            self.requestError(request_id, "Request canceled")

    ##  Emitted whenever data for a request is available.
    #
    #   This will be emitted while a request is processing and requested data has become
    #   available.
    #
    #   \param request_id The ID of the request that data is available for.
    #   \param data A dictionary with virtual paths and data for those paths.
    @dbus.decorators.signal("nl.ultimaker.charon", "sa{sv}")
    def requestData(self, request_id, data):
        pass

    ##  Emitted whenever a request for data has been completed.
    #
    #   This signal will be emitted once a request is completed successfully.
    #
    #   \param request_id The ID of the request that completed.
    @dbus.decorators.signal("nl.ultimaker.charon", "s")
    def requestCompleted(self, request_id):
        pass

    ##  Emitted whenever a request that is processing encounters an error.
    #
    #   \param request_id The ID of the request that encountered an error.
    #   \param error_string A string describing the error.
    @dbus.decorators.signal("nl.ultimaker.charon", "ss")
    def requestError(self, request_id, error_string):
        pass



import enum
import threading
import uuid
from typing import List, Dict, Any, Optional, Callable

from .DBusInterface import DBusInterface


##  Wrapper around the Charon DBus service that hides the DBus details.
#
#   This class encapsulates all the data and information needed for
#   retrieving some data from a file supported by the Charon file service.
#
#   It can be used to simplify dealing with the DBus service.
class Request:
    # The request state.
    class State(enum.IntEnum):
        Initial = 0  # Request was created, but not started yet.
        Running = 1  # Request was started.
        Completed = 2  # Request completed successfully.
        Error = 3  # Request encountered an error.

    ##  Constructor.
    #
    #   \param file_path The path to a file to get data from.
    #   \param virtual_paths A list of virtual paths with the data to retrieve.
    def __init__(self, file_path: str, virtual_paths: List[str]) -> None:
        self.__file_path = file_path
        self.__virtual_paths = virtual_paths

        self.__state = self.State.Initial
        self.__request_id = 0
        self.__data = {} # type: Dict[str, Any]
        self.__error_string = ""

        self.__event = threading.Event()

        self.__request_data_callback = None # type: Optional[Callable[["Request", Dict[str, Any]], None]]
        self.__request_completed_callback = None # type: Optional[Callable[["Request"], None]]
        self.__request_error_callback = None # type: Optional[Callable[["Request", str], None]]

    ##  Cleanup function.
    def __del__(self):
        if self.__state != self.State.Initial:
            self.stop()

            DBusInterface.disconnectSignal("requestData", self.__onRequestData)
            DBusInterface.disconnectSignal("requestCompleted", self.__onRequestCompleted)
            DBusInterface.disconnectSignal("requestError", self.__onRequestError)

    ##  The file path for this request.
    @property
    def filePath(self) -> str:
        return self.__file_path

    ##  The virtual paths for this request.
    @property
    def virtualPaths(self) -> List[str]:
        return self.__virtual_paths

    ##  The state of this request.
    @property
    def state(self) -> State:
        return self.__state

    ##  The data associated with this request.
    #
    #   Note that this will be an empty dictionary until the request
    #   completed.
    @property
    def data(self) -> Dict[str, Any]:
        return self.__data

    ##  A description of the error that was encountered during the request.
    #
    #   Note that this will be an empty string if there was no error.
    @property
    def errorString(self) -> str:
        return self.__error_string

    ##  Set the callbacks that should be called while the request is running.
    #
    #   Note: These parameters can only be passed as keyword arguments.
    #   \param data The callback to call when data is received. Will be passed the request object and a dict with data.
    #   \param completed The callback to call when the request has completed. Will be passed the request object.
    #   \param error The callback to call when the request encountered an error. Will be passed the request object and a string describing the error.
    #
    def setCallbacks(self, *,
            data: Callable[["Request", Dict[str, Any]], None] = None,
            completed: Callable[["Request"], None] = None,
            error: Callable[["Request", str], None] = None) -> None:
        self.__request_data_callback = data
        self.__request_completed_callback = completed
        self.__request_error_callback = error

    ##  Start the request.
    def start(self):
        if self.__state != self.State.Initial:
            return

        self.__request_id = str(uuid.uuid4())

        DBusInterface.connectSignal("requestData", self.__onRequestData)
        DBusInterface.connectSignal("requestCompleted", self.__onRequestCompleted)
        DBusInterface.connectSignal("requestError", self.__onRequestError)

        self.__state = self.State.Running

        DBusInterface.callAsync("startRequest", self.__startSuccess, self.__startError, "ssas", self.__request_id, self.__file_path, self.__virtual_paths)

    ##  Stop the request.
    #
    #   Note that this may fail if the file service was already processing the request.
    def stop(self):
        if self.__state != self.State.Running:
            return

        DBusInterface.callAsync("cancelRequest", None, None, "s", self.__request_id)

    ##  Wait until the request is finished.
    #
    #   Warning! This method will block the calling thread until it is finished. The DBus implementations
    #   require a running event loop for signal delivery to work. This means that if you block the main
    #   loop with this method, you will deadlock since the completed signal is never received.
    def waitForFinished(self):
        if self.__state == self.State.Initial:
            self.start()

        self.__event.clear()
        self.__event.wait()

    def __startSuccess(self, start_success: bool):
        if not start_success:
            self.__startError("Could not start the request")
            return

    def __startError(self, error: str):
        self.__state = self.State.Error
        self.__error_string = error
        self.__event.set()

        if self.__request_error_callback:
            self.__request_error_callback(self, error)

    def __onRequestData(self, request_id: str, data: Dict[str, Any]):
        if self.__state != self.State.Running:
            return

        if self.__request_id != request_id:
            return

        self.__data.update(data)

        if self.__request_data_callback:
            self.__request_data_callback(self, data)

    def __onRequestCompleted(self, request_id: str):
        if self.__state != self.State.Running:
            return

        if self.__request_id != request_id:
            return

        self.__state = self.State.Completed

        if self.__request_completed_callback:
            self.__request_completed_callback(self)

        self.__event.set()

    def __onRequestError(self, request_id: str, error_string: str):
        if self.__request_id != request_id:
            return

        self.__state = self.State.Error
        self.__error_string = error_string

        if self.__request_error_callback:
            self.__request_error_callback(self, error_string)

        self.__event.set()

    def __repr__(self):
        return "<Charon.Client.Request ({id}) file_path = {path} virtual_paths = {virtual} >".format(id = id(self), path = self.__file_path, virtual = self.__virtual_paths)



# Copyright (c) 2018 Ultimaker B.V.	
# libCharon is released under the terms of the LGPLv3 or higher.	

from distutils.core import setup	

setup(	
    name = "Charon",	
    version = "1.0",	
    description = "Library to read and write file packages.",	
    author = "Ultimaker",	
    author_email = "plugins@ultimaker.com",
    url = "https://github.com/Ultimaker/libCharon",	
    packages = ["Charon", "Charon.Client", "Charon.Service", "Charon.filetypes"]	
)



import os
import logging
from typing import Callable, Optional, Union, Any

# We want to use either dbus-python or QtDBus for handling DBus.
# We first need to try importing the module, if that fails we know
# there is no chance of using Qt in the first place. If it succeeds,
# we still may not be using Qt for the main loop, but this check is
# done at runtime.
_has_qt = False
try:
    from PyQt5.QtCore import QCoreApplication, QObject, pyqtSlot
    from PyQt5.QtDBus import QDBusConnection, QDBusMessage, QDBusReply, QDBusInterface, QDBusPendingCallWatcher
    _has_qt = True
except ImportError:
    pass

# Always also try to import dbus-python, since we need to determine things
# at runtime.
try:
    import dbus
    import dbus.mainloop.glib
    from gi.repository import GLib
except ImportError:
    if not _has_qt:
        raise ImportError("Either QtDBus or dbus-python should be available!")

GLib.threads_init()
dbus.mainloop.glib.threads_init()

log = logging.getLogger(__name__)


##  Provides a wrapper around dbus-python or QtDBus to make DBus calls
#
#   Since signals and async method calls are pretty tightly linked to the main
#   loop implementation, we try to use the DBus implementation that matches with
#   the main loop. This class abstracts those details away.
#
#   There are two levels of checks, the first is an import check listed above. The
#   second is a runtime check to see if there is a Qt main loop. If both of those
#   pass, we use QtDBus. If it fails, we use dbus-python.
class DBusInterface:
    # Define default paths that can be used.
    DefaultServicePath = "nl.ultimaker.charon"
    DefaultObjectPath = "/nl/ultimaker/charon"
    DefaultInterface = "nl.ultimaker.charon"

    ##  Make a synchronous call to a DBus method.
    #
    #   \param method_name The name of the method to call.
    #   \param signature The method's argument signature.
    #   \param args Arguments to pass to the DBus method.
    #
    #   The following can only be used as keyword arguments. They default to the
    #   Default* constants defined in this class.
    #
    #   \param service_path The path to the service to call the method on.
    #   \param object_path The object path of the service to call the method on.
    #   \param interface The interface name of the method to call.
    @classmethod
    def callMethod(cls, method_name: str, signature: str, *args, service_path: str = DefaultServicePath, object_path: str = DefaultObjectPath, interface: str = DefaultInterface) -> Any:
        cls.__ensureDBusSetup()
        assert cls.__connection is not None

        if cls.__use_qt:
            message = QDBusMessage.createMethodCall(service_path, object_path, interface, method_name)
            message.setArguments(args)
            result = QDBusReply(cls.__connection.call(message))
            if result.isValid():
                return result.value()
            else:
                log.warning("Did not receive a valid reply for method call %s", method_name)
                log.warning(result.error().message())
                return None

        else:
            return cls.__connection.call_blocking(service_path, object_path, interface, method_name, signature, args)

    ##  Make an asynchronous call to a DBus method.
    #
    #   \param method_name The name of the method to call.
    #   \param success_callback The Callable to call if the method call was successful.
    #   \param error_callback The Callable to call if the method call was unsuccessful.
    #   \param signature The method's argument signature.
    #   \param args Arguments to pass to the DBus method.
    #
    #   The following can only be used as keyword arguments. They default to the
    #   Default* constants defined in this class.
    #
    #   \param service_path The path to the service to call the method on.
    #   \param object_path The object path of the service to call the method on.
    #   \param interface The interface name of the method to call.
    @classmethod
    def callAsync(cls, method_name: str, success_callback: Callable[..., None], error_callback: Callable[..., None], signature: str, *args, service_path: str = DefaultServicePath, object_path: str = DefaultObjectPath, interface: str = DefaultInterface) -> None:
        cls.__ensureDBusSetup()
        assert cls.__connection is not None

        if cls.__use_qt:
            assert cls.__signal_forwarder is not None
            
            message = QDBusMessage.createMethodCall(service_path, object_path, interface, method_name)
            message.setArguments(args)
            cls.__signal_forwarder.asyncCall(message, success_callback, error_callback)
        else:
            cls.__connection.call_async(service_path, object_path, interface, method_name, signature, args, success_callback, error_callback)

    ##  Connect to a DBus signal.
    #
    #   \param signal_name The name of the signal to connect to.
    #   \param callback The callable to call when the signal is received.
    #
    #   The following can only be used as keyword arguments. They default to the
    #   Default* constants defined in this class.
    #
    #   \param service_path The path to the service to call the method on.
    #   \param object_path The object path of the service to call the method on.
    #   \param interface The interface name of the method to call.
    @classmethod
    def connectSignal(cls, signal_name: str, callback: Callable[..., None], *, service_path: str = DefaultServicePath, object_path: str = DefaultObjectPath, interface: str = DefaultInterface) -> bool:
        cls.__ensureDBusSetup()

        if cls.__use_qt:
            assert cls.__signal_forwarder is not None
            return cls.__signal_forwarder.addConnection(service_path, object_path, interface, signal_name, callback)
        else:
            assert cls.__connection is not None
            cls.__connection.add_signal_receiver(callback, signal_name, interface, service_path, object_path)
            return True

    ##  Disconnect from a DBus signal connection.
    #
    #   \param signal_name The name of the signal to disconnect from.
    #   \param callback The Callable to disconnect from the signal.
    #
    #   The following can only be used as keyword arguments. They default to the
    #   Default* constants defined in this class.
    #
    #   \param service_path The path to the service to call the method on.
    #   \param object_path The object path of the service to call the method on.
    #   \param interface The interface name of the method to call.
    @classmethod
    def disconnectSignal(cls, signal_name: str, callback: Callable[..., None], *, service_path: str = DefaultServicePath, object_path: str = DefaultObjectPath, interface: str = DefaultInterface) -> bool:
        cls.__ensureDBusSetup()

        if cls.__use_qt:
            assert cls.__signal_forwarder is not None
            return cls.__signal_forwarder.removeConnection(service_path, object_path, interface, signal_name, callback)
        else:
            assert cls.__connection is not None
            cls.__connection.remove_signal_receiver(callback, signal_name, interface, service_path, object_path)
            return True

    # Private method to ensure we have a DBus connection.
    @classmethod
    def __ensureDBusSetup(cls):
        if cls.__connection:
            return

        if _has_qt and QCoreApplication.instance():
            if os.environ.get("CHARON_USE_SESSION_BUS", 1) == 1:
                cls.__connection = QDBusConnection.sessionBus()
            else:
                cls.__connection = QDBusConnection.systemBus()

            cls.__signal_forwarder = DBusSignalForwarder(cls.__connection)
            cls.__use_qt = True
            return

        if os.environ.get("CHARON_USE_SESSION_BUS", 0) == 1:
            cls.__connection = dbus.Bus.get_session()
        else:
            GLib.MainLoop().run()
            cls.__connection = dbus.SystemBus(private=True, mainloop=dbus.mainloop.glib.DBusGMainLoop())

    __use_qt = False
    __connection = None # type: Optional[Union[dbus.SystemBus]]
    __signal_forwarder = None # type: Optional[DBusSignalForwarder]

if _has_qt:
    ##  Helper class to handle QtDBus signal connections.
    #
    #   QtDBus wants a QObject for its signal connections. Since we do not want
    #   to make Request a QObject, we need to add an intermediary which receives
    #   the signal and calls the appropriate Callable.
    #
    #   In addition, to make it properly handle success/error callbacks for async
    #   method calls, we need to create a QDBusPendingCallWatcher object that we
    #   can listen to. This has the same limitations as QtDBus signals.
    class DBusSignalForwarder(QObject):
        def __init__(self, dbus_connection):
            super().__init__()
            self.__connection = dbus_connection
            self.__connection.registerObject("/" + str(id(self)), self)

            self.__interface_objects = {}
            self.__connected_signals = set()
            self.__callbacks = {}

            self.__pending_async_calls = {}

        ##  Add a signal connection to process.
        def addConnection(self, service_path, object_path, interface, signal_name,  callback):
            connection = (object_path, interface, signal_name)
            if connection not in self.__connected_signals:
                self.__connection.connect(service_path, object_path, interface, signal_name, self.handleSignal)
                self.__connected_signals.add(connection)

            if connection not in self.__callbacks:
                self.__callbacks[connection] = []
            self.__callbacks[connection].append(callback)

        ##  Remove a signal connection.
        def removeConnection(self, service_path, object_path, interface, signal_name, callback):
            connection = (object_path, interface, signal_name)
            if connection not in self.__connected_signals:
                return

            self.__callbacks[connection].remove(callback)

            # Essentially, we do reference counting of the signal here. If the list
            # of connections for the specified signal becomes empty, also remove the
            # signal handler. This prevents us from listening on signals that are
            # not used.
            if not self.__callbacks[connection]:
                self.__connection.disconnect(service_path, object_path, interface, signal_name, self.handleSignal)
                self.__connected_signals.remove(connection)
                del self.__callbacks[connection]

        # Process a signal from DBus.
        @pyqtSlot(QDBusMessage)
        def handleSignal(self, message):
            connection = (message.path(), message.interface(), message.member())
            if connection not in self.__callbacks:
                return

            for callback in self.__callbacks[connection]:
                callback(*message.arguments())

        # Make an asynchronous DBus call. This will trigger __onAsyncCallFinished once it is done.
        def asyncCall(self, message, success_callback, error_callback):
            watcher = QDBusPendingCallWatcher(self.__connection.asyncCall(message))
            watcher.finished.connect(self.__onAsyncCallFinished)
            self.__pending_async_calls[watcher] = (success_callback, error_callback)

        # Handle async call completion.
        @pyqtSlot(QDBusPendingCallWatcher)
        def __onAsyncCallFinished(self, watcher):
            assert watcher in self.__pending_async_calls

            success_callback = self.__pending_async_calls[watcher][0]
            error_callback = self.__pending_async_calls[watcher][1]
            del self.__pending_async_calls[watcher]

            reply = QDBusReply(watcher)
            if reply.isValid():
                if success_callback:
                    success_callback(reply.value())
            else:
                if error_callback:
                    error_callback(reply.error().message())



import sys

import dbus
import dbus.mainloop.glib
from gi.repository import GLib

import Charon.Client

if len(sys.argv) != 2:
    print("Usage: test.py [file]")
    exit(1)

GLib.threads_init()
dbus.mainloop.glib.threads_init()

loop = GLib.MainLoop()
dbus.set_default_main_loop(dbus.mainloop.glib.DBusGMainLoop())

request = Charon.Client.Request(sys.argv[1], ["/Metadata/thumbnail.png"])
request.setCallbacks(completed=lambda request: loop.quit())

request.start()

loop.run()

if request.state == Charon.Client.Request.State.Completed:
    print("Request Complete")
    print(request.data)
elif request.state == Charon.Client.Request.State.Error:
    print("Request Error")
    print(request.errorString)
else:
    print("Request did not finish properly")
    print(request.state)



# Copyright (c) 2018 Ultimaker B.V.
# libCharon is released under the terms of the LGPLv3 or higher.
import os

from Charon.FileInterface import FileInterface  # The interface we're implementing.
from Charon.OpenMode import OpenMode  #To open local files with the selected open mode.
# The supported file types.
from Charon.filetypes.UltimakerFormatPackage import UltimakerFormatPackage
from Charon.filetypes.GCodeFile import GCodeFile
from Charon.filetypes.GCodeGzFile import GCodeGzFile
from Charon.filetypes.GCodeSocket import GCodeSocket

extension_to_mime = {
    ".ufp": "application/x-ufp",
    ".gcode": "text/x-gcode",
    ".gz": "text/x-gcode-gz",
    ".gcode.gz": "text/x-gcode-gz",
    ".gsock": "text/x-gcode-socket"
}

mime_to_implementation = {
    "application/x-ufp": UltimakerFormatPackage,
    "text/x-gcode": GCodeFile,
    "text/x-gcode-gz": GCodeGzFile,
    "text/x-gcode-socket": GCodeSocket
}


##  A facade for a file object.
#
#   This facade finds the correct implementation based on the MIME type of the
#   file it needs to open.
class VirtualFile(FileInterface):
    def __init__(self):
        self._implementation = None

    def open(self, path, mode = OpenMode.ReadOnly, *args, **kwargs):
        _, extension = os.path.splitext(path)
        if extension not in extension_to_mime:
            raise IOError("Unknown extension \"{extension}\".".format(extension = extension))
        mime = extension_to_mime[extension]
        implementation = mime_to_implementation[mime]
        return self.openStream(implementation.stream_handler(path, mode.value + "b"), mime, mode, *args, **kwargs)

    def openStream(self, stream, mime, mode = OpenMode.ReadOnly, *args, **kwargs):
        self._implementation = mime_to_implementation[mime]()
        return self._implementation.openStream(stream, mime, mode, *args, **kwargs)

    def close(self, *args, **kwargs):
        if self._implementation is None:
            raise IOError("Can't close a file before it's opened.")
        result = self._implementation.close(*args, **kwargs)
        self._implementation = None  # You have to open a file again, which might need a different implementation.
        return result

    ##  Causes all calls to functions that aren't defined in this class to be
    #   passed through to the implementation.
    def __getattribute__(self, item):
        if item == "open" or item == "openStream" or item == "close" or item == "__del__" or item == "_implementation":
            # Attributes that VirtualFile overwrites should be called normally.
            return object.__getattribute__(self, item)
        if not object.__getattribute__(self, "_implementation"):
            raise IOError("Can't use '{attribute}' before a file is opened.".format(attribute = item))
        return getattr(self._implementation, item)

    ##  When the object is deleted, close the file.
    def __del__(self):
        if self._implementation is not None:
            self.close()



# Copyright (c) 2018 Ultimaker B.V.
# libCharon is released under the terms of the LGPLv3 or higher.

##  Exception to indicate that an attempt was made to write to a resource that
#   is read-only.
#
#   Normally this sort of thing would be a ``PermissionError`` (the built-in
#   Python exception), but we want to be able to distinguish between these
#   errors ordinary ``PermissionErrors`` raised by the file system not having
#   access to that file.
class ReadOnlyError(PermissionError):
    ##  Creates the exception instance.
    #   \param virtual_path The resource that could not be written to. If not
    #   provided, an empty string is used which indicates that the entire file
    #   could not be written to.
    def __init__(self, virtual_path: str = "") -> None:
        self.virtual_path = virtual_path

    ##  Provides a human-readable version of this error for in the stack trace.
    def __repr__(self) -> str:
        return "ReadOnlyError({resource})".format(resource = self.virtual_path)


# Copyright (c) 2018 Ultimaker B.V.
# Charon is released under the terms of the LGPLv3 or higher.
import io #To create fake streams to write to and read from.
import os #To find the resources with test packages.
import pytest #This module contains unit tests.
import zipfile #To inspect the contents of the zip archives.
import xml.etree.ElementTree as ET #To inspect the contents of the OPC-spec files in the archives.
from collections import OrderedDict
from typing import List, Generator

from Charon.filetypes.OpenPackagingConvention import OpenPackagingConvention, OPCError  # The class we're testing.
from Charon.OpenMode import OpenMode #To open archives.

##  Returns an empty package that you can read from.
#
#   The package has no resources at all, so reading from it will not find
#   anything.
@pytest.fixture()
def empty_read_opc() -> Generator[OpenPackagingConvention, None, None]:
    result = OpenPackagingConvention()
    result.openStream(open(os.path.join(os.path.dirname(__file__), "resources", "empty.opc"), "rb"))
    yield result
    result.close()


##  Returns a package that has a single file in it.
#
#   The file is called "hello.txt" and contains the text "Hello world!" encoded
#   in UTF-8.
@pytest.fixture()
def single_resource_read_opc() -> Generator[OpenPackagingConvention, None, None]:
    result = OpenPackagingConvention()
    result.openStream(open(os.path.join(os.path.dirname(__file__), "resources", "hello.opc"), "rb"))
    yield result
    result.close()


##  Returns an empty package that you can write to.
#
#   Note that you can't really test the output of the write since you don't have
#   the stream it writes to.
@pytest.fixture()
def empty_write_opc() -> Generator[OpenPackagingConvention, None, None]:
    result = OpenPackagingConvention()
    result.openStream(io.BytesIO(), "application/x-opc", OpenMode.WriteOnly)
    yield result
    result.close()


#### Now follow the actual tests. ####

##  Tests whether an empty file is recognised as empty.
def test_listPathsEmpty(empty_read_opc: OpenPackagingConvention):
    assert len(empty_read_opc.listPaths()) == 0


##  Tests getting write streams of various resources that may or may not exist.
#
#   Every test will write some arbitrary data to it to see that that also works.
@pytest.mark.parametrize("virtual_path", ["/dir/file", "/file", "dir/file", "file", "/Metadata"]) #Some extra tests without initial slash to test robustness.
def test_getWriteStream(empty_write_opc: OpenPackagingConvention, virtual_path: str):
    stream = empty_write_opc.getStream(virtual_path)
    stream.write(b"The test is successful.")


##  Tests not allowing to open relationship file directly to prevent mistakes.
@pytest.mark.parametrize("virtual_path", ["/_rels/.rels"])
def test_getWriteStream_forbidOnRels(empty_write_opc: OpenPackagingConvention, virtual_path: str):
    with pytest.raises(OPCError):
        empty_write_opc.getStream(virtual_path)


##  Tests writing data to an archive, then reading it back.
@pytest.mark.parametrize("virtual_path", ["/dir/file", "/file", "/Metadata"]) #Don't try to read .rels back. That won't work.
def test_cycleSetDataGetData(virtual_path: str):
    test_data = b"Let's see if we can read this data back."

    stream = io.BytesIO()
    package = OpenPackagingConvention()
    package.openStream(stream, mode = OpenMode.WriteOnly)
    package.setData({virtual_path: test_data})
    package.close()

    stream.seek(0)
    package = OpenPackagingConvention()
    package.openStream(stream)
    result = package.getData(virtual_path)

    assert len(result) == 1 #This data must be the only data we've found.
    assert virtual_path in result #The path must be in the dictionary.
    assert result[virtual_path] == test_data #The data itself is still correct.


@pytest.mark.parametrize("virtual_path, path_list", [
    ("/foo/materials", ["/foo/materials", "/[Content_Types].xml", "/_rels/.rels"]),
    ("/materials", ["/files/materials", "/[Content_Types].xml", "/_rels/.rels"])
])
def test_aliases_replacement(virtual_path: str, path_list: List[str]):
    test_data = b"Let's see if we can read this data back."

    stream = io.BytesIO()
    package = OpenPackagingConvention()
    package._aliases = OrderedDict([
        (r"/materials", "/files/materials")
    ])
    package.openStream(stream, mode = OpenMode.WriteOnly)
    package.setData({virtual_path: test_data})
    package.close()

    stream.seek(0)
    package = OpenPackagingConvention()
    package.openStream(stream)
    result = package.listPaths()

    assert result == path_list

##  Tests writing data via a stream to an archive, then reading it back via a
#   stream.
@pytest.mark.parametrize("virtual_path", ["/dir/file", "/file", "/Metadata"])
def test_cycleStreamWriteRead(virtual_path: str):
    test_data = b"Softly does the river flow, flow, flow."

    stream = io.BytesIO()
    package = OpenPackagingConvention()
    package.openStream(stream, mode = OpenMode.WriteOnly)
    resource = package.getStream(virtual_path)
    resource.write(test_data)
    package.close()

    stream.seek(0)
    package = OpenPackagingConvention()
    package.openStream(stream)
    resource = package.getStream(virtual_path)
    result = resource.read()

    assert result == test_data


##  Tests setting metadata in an archive, then reading that metadata back.
@pytest.mark.parametrize("virtual_path", ["/Metadata/some/global/setting", "/hello.txt/test", "/also/global/entry"])
def test_cycleSetMetadataGetMetadata(virtual_path: str):
    test_data = "Hasta la vista, baby."

    stream = io.BytesIO()
    package = OpenPackagingConvention()
    package.openStream(stream, mode = OpenMode.WriteOnly)
    package.setData({"/hello.txt": b"Hello world!"}) #Add a file to attach non-global metadata to.
    package.setMetadata({virtual_path: test_data})
    package.close()

    stream.seek(0)
    package = OpenPackagingConvention()
    package.openStream(stream)
    result = package.getMetadata(virtual_path)
    
    prefixed_virtual_path = "/metadata{}".format(virtual_path)

    assert len(result) == 1 #Only one metadata entry was set.
    assert prefixed_virtual_path in result #And it was the correct entry.
    assert result[prefixed_virtual_path] == test_data #With the correct value.


##  Tests toByteArray with its parameters.
#
#   This doesn't test if the bytes are correct, because that is the task of the
#   zipfile module. We merely test that it gets some bytes array and that the
#   offset and size parameters work.
def test_toByteArray(single_resource_read_opc):
    original = single_resource_read_opc.toByteArray()
    original_length = len(original)

    #Even empty zip archives are already 22 bytes, so offsets and sizes of less than that should be okay.
    result = single_resource_read_opc.toByteArray(offset = 10)
    assert len(result) == original_length - 10 #The first 10 bytes have fallen off.

    result = single_resource_read_opc.toByteArray(count = 8)
    assert len(result) == 8 #Limited to size 8.

    result = single_resource_read_opc.toByteArray(offset = 10, count = 8)
    assert len(result) == 8 #Still limited by the size, even though there is an offset.

    result = single_resource_read_opc.toByteArray(count = 999999) #This is a small file, definitely smaller than 1MiB.
    assert len(result) == original_length #Should be limited to the actual file length.


##  Tests toByteArray when loading from a stream.
def test_toByteArrayStream():
    stream = io.BytesIO()
    package = OpenPackagingConvention()
    package.openStream(stream, mode = OpenMode.WriteOnly)
    package.setData({"/hello.txt": b"Hello world!"}) #Add some arbitrary data so that the file size is not trivial regardless of what format is used.
    package.close()

    stream.seek(0)
    package = OpenPackagingConvention()
    package.openStream(stream)
    result = package.toByteArray()

    assert len(result) > 0 #There must be some data in it.


##  Tests whether a content type gets added and that it gets added in the
#   correct location.
def test_addContentType():
    stream = io.BytesIO()
    package = OpenPackagingConvention()
    package.openStream(stream, mode = OpenMode.WriteOnly)
    package.addContentType("lol", "audio/x-laughing")
    package.close()

    stream.seek(0)
    #This time, open as .zip to just inspect the file contents.
    archive = zipfile.ZipFile(stream)
    assert "/[Content_Types].xml" in archive.namelist()
    content_types = archive.open("/[Content_Types].xml").read()
    content_types_element = ET.fromstring(content_types)

    defaults = content_types_element.findall("{http://schemas.openxmlformats.org/package/2006/content-types}Default")
    assert len(defaults) == 2 #We only added one content type, but there must also be the .rels content type.
    for default in defaults:
        assert "Extension" in default.attrib
        assert "ContentType" in default.attrib
        assert default.attrib["Extension"] in ["lol", "rels"]
        if default.attrib["Extension"] == "lol":
            assert default.attrib["ContentType"] == "audio/x-laughing"
        elif default.attrib["Extension"] == "rels":
            assert default.attrib["ContentType"] == "application/vnd.openxmlformats-package.relationships+xml"


##  Tests whether a relation gets added and that it gets saved in the correct
#   location.
def test_addRelation():
    stream = io.BytesIO()
    package = OpenPackagingConvention()
    package.openStream(stream, mode = OpenMode.WriteOnly)
    package.setData({"/whoo.txt": b"Boo", "/whoo.enhanced.txt": b"BOOOO!", "/whoo.enforced.txt": b"BOOOOOOOOOO!"}) #Need 3 files: One base and two that are related.
    package.addRelation("whoo.enhanced.txt", "An enhanced version of it.", "whoo.txt")
    package.addRelation("whoo.enforced.txt", "A greatly enhanced version of it.", "whoo.txt")
    package.close()

    stream.seek(0)
    #This time, open as .zip to just inspect the file contents.
    archive = zipfile.ZipFile(stream)
    assert "/_rels/whoo.txt.rels" in archive.namelist() #It must create a file specifically for whoo.txt
    relations = archive.open("/_rels/whoo.txt.rels").read()
    relations_element = ET.fromstring(relations)

    both_relations = relations_element.findall("{http://schemas.openxmlformats.org/package/2006/relationships}Relationship")
    assert len(both_relations) == 2 #We added two relations.
    for relation in both_relations:
        assert "Id" in relation.attrib
        assert "Target" in relation.attrib
        assert "Type" in relation.attrib
        assert relation.attrib["Target"] == "/whoo.enhanced.txt" or relation.attrib["Target"] == "/whoo.enforced.txt"
        if relation.attrib["Target"] == "/whoo.enhanced.txt":
            assert relation.attrib["Type"] == "An enhanced version of it."
        elif relation.attrib["Target"] == "/whoo.enforced.txt":
            assert relation.attrib["Type"] == "A greatly enhanced version of it."
    assert both_relations[0].attrib["Id"] != both_relations[1].attrib["Id"] #Id must be unique.


##  Tests getting the size of a file.
#
#   This is implemented knowing the contents of single_resource_read_opc.
def test_getMetadataSize(single_resource_read_opc):
    metadata = single_resource_read_opc.getMetadata("/hello.txt/size")
    assert "/metadata/hello.txt/size" in metadata
    assert metadata["/metadata/hello.txt/size"] == len("Hello world!\n".encode("UTF-8")) #Compare with the length of the file's contents as encoded in UTF-8.



import queue
import threading
import logging
import dbus
from typing import List, Dict, Any

import FileService

import Charon.VirtualFile
import Charon.OpenMode

log = logging.getLogger(__name__)

##  A request for data that needs to be processed.
#
#   Each request will be processed by a worker thread to actually perform the data
#   retrieval.
class Request:
    ##  Constructor.
    #
    #   \param file_service The main FileService object. Used to emit signals.
    #   \param request_id The ID used to identify this request.
    #   \param file_path A path to a file to retrieve data from.
    #   \param virtual_paths The virtual paths to retrieve for this request.
    def __init__(self, file_service: FileService.FileService, request_id: str, file_path: str, virtual_paths: List[str]) -> None:
        self.file_service = file_service
        self.file_path = file_path
        self.virtual_paths = virtual_paths
        self.request_id = request_id

        # This is used a workaround for limitations of Python's Queue class.
        # Queue does not implement a "remove arbitrary item" method. So instead,
        # keep a removed request in the queue and set this flag to true, after
        # which a worker thread can dispose of the object when it encounters
        # the request.
        self.should_remove = False

    ##  Perform the actual data retrieval.
    #
    #   This is a potentially long-running operation that should be handled by a
    #   thread.
    def run(self):
        try:
            virtual_file = Charon.VirtualFile.VirtualFile()
            virtual_file.open(self.file_path, Charon.OpenMode.OpenMode.ReadOnly)

            for path in self.virtual_paths:
                data = virtual_file.getData(path)


                for key, value in data.items():
                    if isinstance(value, bytes):
                        data[key] = dbus.ByteArray(value)

                # dbus-python is stupid and we need to convert the entire nested dictionary
                # into something it understands.
                data = self._convertDictionary(data)

                self.file_service.requestData(self.request_id, data)

            virtual_file.close()
            self.file_service.requestCompleted(self.request_id)
        except Exception as e:
            log.log(logging.DEBUG, "", exc_info = 1)
            self.file_service.requestError(self.request_id, str(e))

    # Helper for dbus-python to convert a nested dict to a nested dict.
    #
    # Yes, really, apparently dbus-python does some really stupid things with dictionaries
    # making this necessary.
    def _convertDictionary(self, dictionary: Dict[str, Any]) -> dbus.Dictionary:
        result = dbus.Dictionary({}, signature = "sv")

        for key, value in dictionary.items():
            key = str(key) # Since we are sending a dict of str, Any, make sure the keys are strings.
            if isinstance(value, bytes):
                # Workaround dbus-python being stupid and not realizing that a bytes object
                # should be sent as byte array, not as string.
                result[key] = dbus.ByteArray(value)
            elif isinstance(value, dict):
                result[key] = self._convertDictionary(value)
            else:
                result[key] = value

        return result

##  A queue of requests that need to be processed.
#
#   This class will maintain a queue of requests to process along with the worker threads
#   to process them. It processes the request in LIFO order.
class RequestQueue:
    def __init__(self):
        self.__queue = queue.LifoQueue(self.__maximum_queue_size)

        # This map is used to keep track of which requests we already received.
        # This is mostly intended to be able to cancel requests that are
        # in the queue.
        self.__request_map = {}

        self.__workers = []

        for i in range(self.__worker_count):
            worker = threading.Thread(target = self.__worker_thread_run, daemon = True)
            worker.start()
            self.__workers.append(worker)

    ##  Add a new request to the queue.
    #
    #   \param request The request to add.
    #
    #   \return True if successful, False if the request could not be enqueued for some reason.
    def enqueue(self, request: Request):
        if(request.request_id in self.__request_map):
            log.debug("Tried to enqueue a request with ID {id} which is already in the queue".format(id = request.request_id))
            return False

        try:
            self.__queue.put(request, block = False)
        except queue.Full:
            log.debug("Tried to enqueue a request with ID {id} but the queue is full".format(id = request.request_id))
            return False

        self.__request_map[request.request_id] = request
        return True

    ##  Remove a request from the queue.
    #
    #   \param request_id The ID of the request to remove.
    #
    #   \return True if the request was successfully removed, False if the request was not in the queue.
    def dequeue(self, request_id: str):
        if request_id not in self.__request_map:
            log.debug("Unable to remove request with ID {id} which is not in the queue".format(id = request_id))
            return False

        self.__request_map[request_id].should_remove = True
        return True

    ##  Take the next request off the queue.
    #
    #   Note that this method will block if there are no current requests on the queue.
    #
    #   \return The next request on the queue.
    def takeNext(self) -> Request:
        request = self.__queue.get()
        del self.__request_map[request.request_id]
        return request

    # Implementation of the worker thread run method.
    def __worker_thread_run(self):
        while True:
            request = self.takeNext()
            if request.should_remove:
                continue

            try:
                request.run()
            except Exception as e:
                log.log(logging.DEBUG, "Request caused an uncaught exception when running!", exc_info = 1)

    __maximum_queue_size = 100
    __worker_count = 2
